/* JPG: 1.9.91 */
/* determinate unfolding with user-controlled unfolding choices */
/* special rule for the Poole-Goebel prover */

resultant(A,R) :-
	resultant(A,_,[],R).

resultant(A,Path,R) :-
	resultant(A,Path,[],R).

resultant(G,Abs,P3,R2) :-
	ref_clause(G,B,N),
	prefix(B,P,P2,R),
	functor(G,F,_),
	unfold_choices(R,F,P2,P3,R2),
	repr(G,Abs,N,P).

repr(member(X,_),T,_,_) :-
	copyterm(member(X,_),T),
	canonical(T),
	!.
repr(member(X,_,_),T,_,_) :-
	copyterm(member(X,_,_),T),
	canonical(T),
	!.
repr(_,[N|P],N,P).
	
/* Determinate unfolding only */

prefix(true,Ns,Ns,true) :-
	!.
prefix(Gs,Ns,Ns1,Gs1) :-
	select_lit(Gs,B,Gs3,N),
	\+ B = member(_,_),
	choices(B,Ch,Gs3),
	Ch=[(C,(H :- R))],
	!,
	replace_goal(Gs,N,Gs2,H,R),
	prefix(Gs2,Ns,[(N,C)|Ns1],Gs1).
prefix(Gs,Ns,Ns,Gs) :-
	no_Deadend(Gs).

select_lit((G,Gs),G,Gs,1).
select_lit((B,Gs),G,(B,Gs1),N1) :-
	!,
	select_lit(Gs,G,Gs1,N),
	N1 is N+1.
select_lit(G,G,true,1).

unfold_choices(true,_,Ns,Ns,true) :-
	!.
unfold_choices(Gs,F,Ns,Ns1,Gs1) :-
	safe_select(Gs,B,N),
	unf(B),
	\+ imported(B),
	!,
	ref_clause(B,R,C),
	replace_goal(Gs,N,Gs2,B,R),
	no_Deadend(Gs2),
	unfold_choices(Gs2,F,Ns,[(N,C)|Ns1],Gs1).
unfold_choices(Gs,F,Ns,Ns1,Gs1) :-
	safe_select(Gs,B,N),
	evaluable(B),
	!,
	sp_solve(B),
	pattern(B,C),
	replace_goal(Gs,N,Gs2,B,true),
	no_Deadend(Gs2),
	unfold_choices(Gs2,F,Ns,[(N,C)|Ns1],Gs1).
unfold_choices(Gs,F,Ns,Ns1,Gs1) :-
	unfold_suffix(Gs,F,Ns,Ns1,Gs1).

unfold_suffix(true,_,Ns,Ns,true) :-
	!.
unfold_suffix(Gs,F,Ns,Ns1,Gs1) :-
        select_lit(Gs,B,Gs3,N),
	\+ B = member(_,_),
	functor(B,F1,_),
	F1 \== F,
        choices(B,Ch,Gs3),
        Ch=[(C,(H :- R))],
        !,
        replace_goal(Gs,N,Gs2,H,R),
        unfold_suffix(Gs2,F,Ns,[(N,C)|Ns1],Gs1).
unfold_suffix(Gs,_,Ns,Ns,Gs) :-
	no_Deadend(Gs).

safe_select((G,_),G,1).
safe_select((B,Gs),G,N1) :-
	sp_builtin(B),
	!,
	safe_select(Gs,G,N),
	N1 is N+1.
safe_select(G,G,1).

replace_goal((G,true),2,G,_,_) :- /* eliminate true at end of goal */
	!.
replace_goal((G,Gs),1,Gs1,G,R) :-
	!,
	joingoals(R,Gs,Gs1).
replace_goal((G,Gs),N,(G,Gs1),H,R) :-
	!,
	N1 is N-1,
	replace_goal(Gs,N1,Gs1,H,R).
replace_goal(G,1,R,G,R).


choices(clause(G,B),Cs,Gs) :-
	!,
	bagof((N,(clause(G,B) :- true)),[Gs]^(ref_clause(G,B,N),
		no_Deadend(Gs)),Cs).
choices(G,Ps,Gs) :-
	sp_builtin(G),
	unfoldable(G),
	!,
	pattern(G,P),
	bagof((P,(G :- true)),[Gs]^(call(G)), Ps).
choices(G,[],_) :-
	sp_builtin(G),
	!.
choices(G,[],_) :-
	imported(G),
	!.
choices(G,Ps,Gs) :-
	evaluable(G),
	!,
	pattern(G,P),
	bagof((P,(G :- true)),[Gs]^(sp_solve(G)), Ps).
choices(G,Cs,Gs) :-
	bagof((N,(G :- B)),
		[Gs]^(ref_clause(G,B,N)), Cs).

/*
choices(clause(G,B),Cs,Gs) :-
	!,
	bagof((N,(clause(G,B) :- true)),[Gs]^(ref_clause(G,B,N),no_Deadend(B),
		no_Deadend(Gs)),Cs).
choices(G,Ps,Gs) :-
	sp_builtin(G),
	unfoldable(G),
	!,
	pattern(G,P),
	bagof((P,(G :- true)),[Gs]^(call(G),no_Deadend(Gs)), Ps).
choices(G,[],_) :-
	sp_builtin(G),
	!.
choices(G,[],_) :-
	imported(G),
	!.
choices(G,Ps,Gs) :-
	evaluable(G),
	!,
	pattern(G,P),
	bagof((P,(G :- true)),[Gs]^(sp_solve(G),no_Deadend(Gs)), Ps).
choices(G,Cs,Gs) :-
	bagof((N,(G :- B)),
		[Gs]^(ref_clause(G,B,N),no_Deadend(B),no_Deadend(Gs)), Cs).
*/
