/* Look for creation of duplicate structures in clause bodies */
/* Factor out duplicate structures and add unification goals  */
/* at start of clause */

/* JPG - 18.9.90  */

cleanup(_,Cls,Cls1) :-
	remove_dup_structs(Cls,Cls1).

remove_dup_structs([],[]) :-
	!.
remove_dup_structs([(H :- B)|Cls],[(H :- B1)|Cls1]) :-
	!,
	dup_structs(B,S),
	add_unif_goals(S,B,B1),
	remove_dup_structs(Cls,Cls1).
remove_dup_structs([Proc|Ps],[Proc1|Ps1]) :-
	remove_dup_structs(Proc,Proc1),
	remove_dup_structs(Ps,Ps1).

dup_structs(B,S) :-
	structs(B,S1,[]),
	duplicates(S1,S).

structs(B,S,S) :-
	var(B),
	!.
structs(B,S,S1) :-
	prolog_meta(B,_,P,_),
	!,
	structs(P,S,S1).
structs((G,Gs),S,S2) :-
	!,
	structs(G,S,S1),
	structs(Gs,S1,S2).
structs(B,S,S1) :-
	B =.. [_|Args],
	arg_structs(Args,S,S1).

arg_structs([],S,S) :-
	!.
arg_structs([X|Xs],[X|S],S1) :-
	nonvar(X),
	functor(X,_,N),
	N > 0,
	!,
	arg_structs(Xs,S,S1).
arg_structs([_|Xs],S,S1) :-
	arg_structs(Xs,S,S1).

duplicates([],[]) :-
	!.
duplicates([_],[]) :-
	!.
duplicates(S,Ds) :-
	dups(S,Ds,[]).

dups([X|Xs],Dups,D1) :-
	memb3(X,Xs),
	\+ memb3(X,D1),
	!,
	dups(Xs,Dups,[X|D1]).
dups([_|Xs],Dups,D1) :-
	!,
	dups(Xs,Dups,D1).
dups([],D,D).

memb3(X,[X1|_]) :-
	X == X1,
	!.
memb3(X,[_|Xs]) :-
	memb3(X,Xs).

add_unif_goals([],B,B) :-
	!.
add_unif_goals(Dups,B,B1) :-
	unif_goals(Dups,Us),
	replaceargs(Us,B,B2),
	append_unif_goals(Us,B2,B1).

replaceargs(_,B,B) :-
	var(B),
	!.
replaceargs(Us,B1,B2) :-
	prolog_meta(B1,B2,P1,P2),
	!,
	replaceargs(Us,P1,P2).
replaceargs(Us,(G,Gs),(G1,Gs1)) :-
	!,
	replaceargs(Us,G,G1),
	replaceargs(Us,Gs,Gs1).
replaceargs(Us,G,G1) :-
	G =.. [P|Args],
	replace_each_arg(Us,Args,Args1),
	G1 =.. [P|Args1].

replace_each_arg(_,[],[]) :-
	!.
replace_each_arg(Us,[A|As],[X|Xs]) :-
	nonvar(A),
	functor(A,_,N),
	N > 0,
	memb4((X=A),Us),
	!,
	replace_each_arg(Us,As,Xs).
replace_each_arg(Us,[A|As],[A|Xs]) :-
	replace_each_arg(Us,As,Xs).
	
unif_goals([],[]) :-
	!.
unif_goals([A|As],[(_ = A)|Us]) :-
	unif_goals(As,Us).

memb4((X=A),[(X=A1)|_]) :-
	A == A1,
	!.
memb4(X,[_|Xs]) :-
	memb4(X,Xs).

append_unif_goals([],B,B) :-
	!.
append_unif_goals([U|Us],B,(U,B1)) :-
	append_unif_goals(Us,B,B1).

