/* sp.pl */
/* Construct specialised program */
/* 
	1.  Flow analysis performed by abint/3
	2.  Renaming of predicates performed by rename_preds/4
	3.  Construction of new procedures by make_clauses/5
	4.  Writing out to file by write_to_file/1.

*/
/* This core procedure can be adapted to perform
   various kinds of specialisation.  Different abstraction
   operations, and unfolding transformations can be defined.
   The following procedures define variants of the program:

	resultant/2
	more_specific/1
	cleanup/3

 */

/* John Gallagher: 31.7.90 */
/*  revised 31.8.90 */

%:- setdebug.

%:- dynamic directive/1.

sp(G) :-
	vars(G,Xs),
	anybody(Xs,U),
	%tell(temptrace),
	flow_analyse(catom(G,(U,[])),Pfg,Pathtable,0),
	%told,
	rename_preds(catom(G,(U,[])),Pfg,Pathtable,R),
	!,
	make_clauses(catom(G,(U,[])),R,Pfg,Pathtable,Cls),
	!,
	write_to_file(Cls),
        !.

anybody([],true) :-
	!.
anybody([X],any(X)) :-
	!.
anybody([X|Xs],(any(X),As)) :-
	anybody(Xs,As).

flow_analyse(G,Pfg,Pathtable,K0) :-
        nl,
        write('flow analysis---'),sp_ttyflush,
        abint(G,Pfg,Pathtable,K0).

rename_preds(G,Pfg,Pathtable,R) :-
	rep(G,Initrep,Pathtable,Pfg),
	G = catom(G1,_),
	functor(G1,P,_),
	R = [name(Initrep,0,P)|_],
	renames(Pfg,R).

make_clauses(G,R,Pfg,Pathtable,Cls) :-
        nl,write('constructing specialised program---'),sp_ttyflush,
        tell(tempwrite),procedures(R,R,Cls1,Pathtable,Pfg),told,
        nl,nl,write('cleaning up---'),sp_ttyflush,
        cleanup(G,Cls1,Cls2),
        cleanup(G,Cls2,Cls).

write_to_file(Cls) :-
        nl,write(user_output,'Enter filename: '),
        read(user_input,F),
        sp_filewrite(F,Cls).

/* Rename each atom in the environment */

renames([],R) :-
	!,
	list1(R).
renames([element(_,As)|Cs],R) :-
	most_general(As,C),
	renamepred(C,R),
	renames(Cs,R).
	
renamepred(P,S) :-
	renamepred(P,S,0).

renamepred(catom(A,Cs),S,M) :-
	var(S),
	!,
	N is M + 1,
	A =.. [P|_],
	name(P,Plist),
	name(N,Nlist),
	my_append(Plist,[95|Nlist],PNlist), /* 95 = underscore */
	name(Pn,PNlist),
	S = [name(catom(A,Cs),N,Pn)|_],
	!.
renamepred(P,[name(P,_,_)|_],_) :-
	!.
renamepred(P,[name(Q,M,_)|S],_) :-
	P =.. [A|_],
	Q =.. [A|_],
	!,
	renamepred(P,S,M).
renamepred(P,[_|S],K) :-
	renamepred(P,S,K).

/*  Generate procedures for the calls  */

procedures([],_,[],_,_) :-
	!.
procedures([name(A,N,Pn)|Rs],R,[Cls|Prs],Env,Pfg) :-
      procedure(A,Env,R,N,Pn,Cls,Pfg),
	!,
      procedures(Rs,R,Prs,Env,Pfg).

procedure(catom(A,(U,C)),Env,Rs,K,Pn,Cls,Pfg) :- 
	!,
	nl,nl,
	write(A),
	write(' :- '),
	nl,
	write('      '),
	write(U),
	write('.'), 
	nl,
	displayqTypes(C),
	retrieve_resultants(catom(A,(U,C)),Env,Cls1),
	rename_bodies(A,Cls1,Env,Pfg,Rs,Cls,K,Pn).

/*  Retrieve the set of resultants for a given atom */

retrieve_resultants(A,Pathtable,Rs) :-
	get_path(A,Pathtable,_,Rs,_,_).


rename_bodies(_,[],_,_,_,[],_,_) :-
	!.
rename_bodies(A,[(Resultant,_,As)|Cls1],Env,Pfg,Rs,[(H :- R2)|Cls],0,_) :-
	melt(A,H1),
	melt(Resultant,(H1 :- R1,_)),
	H1 =.. [P,_|Args],
	H =.. [P|Args],
	make_body(R1,As,Env,Pfg,Rs,R2),
	!,
	rename_bodies(A,Cls1,Env,Pfg,Rs,Cls,0,_).
rename_bodies(A,[(Resultant,_,As)|Cls1],Env,Pfg,Rs,[(H :- R2)|Cls],K,Pn) :-
	K > 0,
	melt(A,H1),
	H1 =.. [_,_|Args],
	vars(Args,Xs),
	melt(Resultant,(H1 :- R1,_)),
	H =.. [Pn|Xs],
	make_body(R1,As,Env,Pfg,Rs,R2),
	!,
	rename_bodies(A,Cls1,Env,Pfg,Rs,Cls,K,Pn).
rename_bodies(A,[_|Cls1],Env,Pfg,Rs,Cls,K,Pn) :-  	/* couldn't make body */
	rename_bodies(A,Cls1,Env,Pfg,Rs,Cls,K,Pn).	/* because it fails   */

make_body(B,_,_,_,_,B) :- 
	var(B),
	!.
make_body((B,Bs),[D|Ds],Env,Pfg,Rs,(A,As)) :- 
	!,
	buildNewGoal(B,D,Env,Pfg,Rs,A),
	make_body(Bs,Ds,Env,Pfg,Rs,As).
%make_body(B,Ds,Env,Pfg,Rs,B1) :- 
%	prolog_meta(B,B1,P,P1),
%	!,
%	make_body(P,Ds,Env,Pfg,Rs,P1).
make_body(B,D,Env,Pfg,Rs,A) :-
	buildNewGoal(B,D,Env,Pfg,Rs,A).

buildNewGoal(G,builtin,_,_,_,G) :-
	sp_builtin(G),
	!.
buildNewGoal(true,true,_,_,_,true) :-
	!.
buildNewGoal(G,imported,_,_,_,G) :-
	imported(G),
	!.
buildNewGoal(B,catom(A,C),Env,Pfg,Rs,B1) :-
	!,
      rep(catom(A,C),R,Env,Pfg),
      renameGoal(R,B,B1,Rs).
buildNewGoal(B,[D],Env,Pfg,Rs,B1) :-
	buildNewGoal(B,D,Env,Pfg,Rs,B1).


renameGoal(fail,_,fail,_) :-
	!.
renameGoal(catom(R,C),B,B1,Rs) :-
	memb2(name(catom(R,C),0,_),Rs),
	!,
	B =.. [P,_Trace|Args],
	B1 =.. [P|Args].
renameGoal(catom(R,C),B,A,Rs) :-
	memb2(name(catom(R,C),_,Pn),Rs),
	melt(R,Mr),
	Mr =.. [_,_Trace|Args],
	vars(Args,Vs),
	A =.. [Pn|Vs],
	B = Mr,
	!.
renameGoal(_,B,_,_) :-
	nl,
	write(B),
	write('cannot be renamed.......'),
	nl,nl,
	fail.

 
most_general([A|As],A) :-
	var(As),
	!.
most_general([_|As],A) :-
	most_general(As,A).

rep(A,B,Pathtable,Pfg) :-
	get_path(A,Pathtable,P,_,_,_),
	get_reps(P,Pfg,Bs),
	most_general(Bs,B),
	!.
rep(_,fail,_,_) :-
	nl,
	write('Failed atom detected:  replaced by ''fail'''),
	nl.

get_reps(P,[element(P1,Bs)|_],Bs) :-
	P == P1,
	!.
get_reps(P,[_|Es],Bs) :-
	get_reps(P,Es,Bs).

sp_filewrite(tty,Cls) :-
        !,
	nl,write('Specialised Program:'),nl,nl,
        displayqProcs(Cls),
	nl,nl.
sp_filewrite(F,Cls) :-
        tell(F),
	write_directives,
        displayqProcs(Cls),
        told.

/* write the program to current output file  */

displayqProcs([P|Ps]) :-
        displayqClauses(P),
        displayqProcs(Ps).
displayqProcs([]).

write_directives :-
	bagof(X,(directive(X),X \== dummy,write_directive(X)),_),
	!.
write_directives.

write_directive(X) :-
	canonical(X),
	write(':- '),
	write(X),
	write('.'),
	nl.

imported(X) :- ub_clause(imported(_,X),true).

cleanup(_,P,P).

