% computes a query-answer transform with three phases:
%	1.  bottom-up answers
%	2.  queries using b-up answers (as in Jacobs-Langen)
%	3.  answers restricted by queries

ans_clauses(X) :- 
	end_of_file(X),
	!.
ans_clauses((A :- B)) :- 
	!,
	canonical((A :- B)),
	answerlits(B,Bs),
	flat_name('_ans'(A),AA),
	flat_name(query(A),QA),
	flat_names(Bs,FBs),
	displayqRClauses([(AA :- QA,FBs)]),
	read(C1),
	!,
	ans_clauses(C1).
ans_clauses(A) :- 
        canonical(A),
	flat_name('_ans'(A),AA),
        flat_name(query(A),QA),
        displayqRClauses([(AA :- QA)]),
	read(C1),
	!,
	ans_clauses(C1).

answerlits((B,Bs),(B,Bs1)) :-
	sp_builtin(B),
	!,
	answerlits(Bs,Bs1).
answerlits((B,Bs),('_ans'(B),Bs1)) :-
	!,
	answerlits(Bs,Bs1).
answerlits(B,B) :-
	sp_builtin(B).
answerlits(B,'_ans'(B)).

flat_name(A,A) :-
	sp_builtin(A),
	!.
flat_name(A,A1) :-
	A =.. [F,As],
	As =.. [G|Args],
	name(F,Fn),
	name(G,Gn),
	concat(Gn,[95|Fn],An),
	name(H,An),
	A1 =.. [H|Args].

flat_names((B,Bs),(B1,Bs1)) :-
	!,
	flat_name(B,B1),
	flat_names(Bs,Bs1).
flat_names(B,B1) :-
	flat_name(B,B1).

magic(F,F1,Q) :-
	see(F),
	tell(F1),
	read(C),
	ans_clauses(C),
	seen,
	see(F),
	read(C1),
	query_clauses(C1),
	flat_name(query(Q),QQ),
	canonical( QQ ),
	displayqRClauses([(QQ :- true)]),
	nl,
	seen,
	see(F),
	read(C2),
	orig_clauses(C2),
	nl,
	seen,
	told,
	new_line.

query_clauses(Z) :- 
	end_of_file(Z),
	!.
query_clauses((A :- B)) :- 
	!,
	canonical((A :- B)),
	bodylitqueries(B,1,A,B),
	read(C1),
	!,
	query_clauses(C1).
query_clauses(_) :- 
	read(C1),
	!,
	query_clauses(C1).

bodylitqueries(true,_,_,_) :-
	!.
bodylitqueries((Bj,Bs),J,A,Lits) :-
	sp_builtin(Bj),
	!,
	J1 is J+1,
	bodylitqueries(Bs,J1,A,Lits).
bodylitqueries((Bj,Bs),J,A,Lits) :-
	!,
	remove_jth(J,Lits,Bj,Lits1),
	flat_name(query(Bj),QBj),
	flat_name(query(A),QA),
	displayqRClauses([(QBj :- QA,Lits1)]),
	J1 is J+1,
	bodylitqueries(Bs,J1,A,Lits).
bodylitqueries(Bj,_,_,_) :-
	sp_builtin(Bj),
	!.
bodylitqueries(Bj,J,A,Lits) :-
	!,
	remove_jth(J,Lits,Bj,Lits1),
	answerlits(Lits1,Bs2),
	flat_names(Bs2,QBs2),
	flat_name(query(Bj),QBj),
	flat_name(query(A),QA),
	displayqRClauses([(QBj :- QA,QBs2)]).

concat([],Xs,Xs).
concat([X|Xs],Ys,[X|Zs]) :-
	concat(Xs,Ys,Zs).


orig_clauses(X) :- 
	end_of_file(X),
	!.
orig_clauses((A :- B)) :- 
	!,
	canonical((A :- B)),
	displayqRClauses([(A :- B)]),
	read(C1),
	!,
	orig_clauses(C1).
orig_clauses(A) :- 
        canonical((A :- true)),
        displayqRClauses([(A :- true)]),
	read(C1),
	!,
	orig_clauses(C1).

remove_jth(1,(B,Bs),B,Bs) :-
        !.
remove_jth(1,B,B,true) :-
	!.
remove_jth(J,(B,Bs),B1,(B,Bs1)) :-
        J1 is J-1,
        remove_jth(J1,Bs,B1,Bs1).


