% JPG - 14.2.92
% Compute an abstraction of a program using the domain Prop 

%:- setdebug.
%:- dynamic user_proc/3.
%:- alldynamic.
 
tp_fix(P) :-
	time(tp_fix1(P,F)),
	display_fix(F).

tp_fix1(P,F) :-
	strong_connected_components(P,G),% compute the dependency ordering
	tp_fix_groups(G,[],F).

tp_fix_groups([],F,F) :-
	!.
tp_fix_groups([(R,G)|Gs],F,F2) :-
	nonrec(R),
%	write(G),
%	nl,
	!,
	init(G,R0),
	tp_alpha(G,[[]],F,R0,R2),  % no need to iterate if non-rec.
	app(F,R2,F1),
	tp_fix_groups(Gs,F1,F2).
tp_fix_groups([(_,G)|Gs],F,F2) :-
%	write(G),nl,
	init(G,R0),
	tp_iterate([[]],G,F,R0,R2),
	app(F,R2,F1),
	tp_fix_groups(Gs,F1,F2).

nonrec(non_recursive).

recursive_pred(X) :-
	memb2(P,X),
	recursive(P),
	!.

init([_|Ps],[[]|R]) :-
	init(Ps,R).
init([],[]).

tp_iterate([],_,_,R,R) :-
	!.
tp_iterate(Nofix,Ps,Acc,R1,R) :-
	% writestate(R1),
	app(R1,Acc,D),
	tp_alpha(Ps,Nofix,D,R1,R2),
	% countdefs(R2),
	check_fix(Ps,R1,R2,Fix),
	!,
	tp_iterate(Fix,Ps,Acc,R2,R).

countdefs([R|Rs]) :-
	sp_length(R,N),write(N),write(' '),
	!,
	countdefs(Rs).
countdefs([]) :-
	nl.


	
	
tp_alpha([],_,_,[],[]).
tp_alpha([P/N|Ps],Nofix,R,[PR|PRs],[UP|R1]) :-
	user_clauses(P/N,Cls),
	infer_proc(Cls,Nofix,R,As),
	ub_proc(As,PR,UP),
	tp_alpha(Ps,Nofix,R,PRs,R1).

infer_proc([(_ :- B)|Cls],Nofix,R,As) :- 	% if no predicate in the body
	nochange(B,Nofix),			% changed, skip clause
	!,
	infer_proc(Cls,Nofix,R,As).
infer_proc([(H :- B)|Cls],Nofix,R,[S|As]) :-
	tsolve(B,R,T),
	unfold_body(T,T1),
	!,
	headtype(H,T1,S), 	
	infer_proc(Cls,Nofix,R,As).
infer_proc([_|Cls],Nofix,R,As) :-		% if no solution in R.
	infer_proc(Cls,Nofix,R,As).
infer_proc([],_,_,[]).

nochange(_,[[]]) :-  	% everything handled on 1st iteration
	!,
	fail.
nochange(true,_) :-
	!.
nochange(B,Nofix) :-
	preds(B,Ps),
	\+ commonpred(Ps,Nofix).

preds(true,[]) :-
	!.
preds((B,Bs), [P/N|Ps]) :-
	!,
	functor(B,P,N),
        preds(Bs,Ps).
preds(B,[P/N]) :-
	functor(B,P,N).

commonpred(X,Y) :-
	memb2(Z,X),
	memb2(Z,Y).
	
tsolve(true,_,true) :-
	!.
tsolve((B,Bs),R,T) :-
	!,
	tsolve(B,R,T1),
	tsolve(Bs,R,T2),
	app(T1,T2,T).
tsolve(B,_,true) :-
        sp_builtin(B),
        !.
tsolve(B,R,T1) :-
	functor(B,S,N),
	headclause(S/N,R,(H :- T)),
	melt((H :- T),(B :- T1)).

unfold_body([],[]) :-
	!.
unfold_body([B|Bs],C) :-
	!,
	unfold_body(B,C1),
	unfold_body(Bs,C2),
	app(C1,C2,C).
unfold_body(implies(X,Y),W) :-
	!,
	vars(X,U),
	vars(Y,V),
	unfold_implies(U,V,W).
unfold_body(X,U) :-
	vars(X,U).

unfold_implies([],V,V) :-
	!.
unfold_implies(_,[],[]) :-
	!.
unfold_implies(U,V,[implies(U,V)]).

headtype(T, T1, (X :- Bs)) :- 
	T =.. [P|Ts],
	functor(T,P,N),
	functor(X,P,N),
	X =.. [P|Xs],
	imply_pairs(Xs,Ts,Pairs),
	app(Pairs,T1,T2),
	headtype1(Xs,Ts,Xs,T2,Bs),
	canonical((X :- Bs)).
	
imply_pairs([],[],[]).
imply_pairs([X|Xs],[T|Ts],[implies([X],Tv)|Ps]) :-
	vars(T,Tv),
	imply_pairs(Xs,Ts,Ps).

headtype1([],[],_,_,[]) :-
	!.
headtype1([X|Xs],[T|Ts],Ys,B1,[S|B]) :-
	vars(T,Tv),
	elim_local_vars(X,Tv,Ys,B1,S),
	!,
	headtype1(Xs,Ts,Ys,B1,B).

elim_local_vars(X,Tv,Ys,B1,implies(Ls,[X])) :-
	elim_locals(X,Tv,Ys,B1,Ls,[]),
	!.
elim_local_vars(_,_,_,_,[]).

elim_locals(_,[],_,_,[],_).
elim_locals(X,[K|Ks],Ys,B,Ls,S) :-
	elim_local_var(X,K,Ys,B,Ls1,S),
	elim_locals(X,Ks,Ys,B,Ls2,S),
	var_union(Ls1,Ls2,Ls).

elim_local_var(_,K,Ys,_,[K],_) :-
	memb3(K,Ys),
	!.
elim_local_var(X,K,Ys,B,S,Path) :-
	no_loop(K,Path),
	!,
	reduce_local(K,B,X,Vs),
	elim_locals(X,Vs,Ys,B,S,[K|Path]).
	
no_loop(K,L) :-
	\+ memb3(K,L).

reduce_local(X,[Y|_],_,[]) :-
	X == Y,
	!.
reduce_local(X,[implies(K,M)|_],W,K) :-
	nonvar(K),
	K \== [W],
	memb3(X,M).
reduce_local(X,[_|B],Y,K) :-
	reduce_local(X,B,Y,K).

check_fix([],[],[],[]).
check_fix([_|Ps],[RP|R],[RP1|S],Qs) :-
        fixpoint(RP,RP1),
	!,
        check_fix(Ps,R,S,Qs).
check_fix([P|Ps],[_|R],[_|S],[P|Qs]) :-
        check_fix(Ps,R,S,Qs).

fixpoint(X,X).
	
ub_proc([],[],[]).
ub_proc([(S :- B)|As],[],UP) :-
	prop_sharing(S,[(S :- B)|As],UP).
ub_proc(As,(S :- B1),UP) :-
	prop_sharing(S,[(S :- B1)|As],UP).

prop_sharing(S,Ps,(S :- Q)) :-
	functor(S,_,N),
	each_arg_sharing(S,1,N,Ps,Q).

each_arg_sharing(_,M,N,_,[]) :-
	M > N,
	!.
each_arg_sharing(S,J,N,Ps,[implies(K,[X])|Q]) :-
	arg(J,S,X),
	all_args(X,Ps,K),
	!,
	J1 is J+1,
	each_arg_sharing(S,J1,N,Ps,Q).
each_arg_sharing(S,J,N,Ps,[[]|Q]) :-
	J1 is J+1,
	each_arg_sharing(S,J1,N,Ps,Q).

all_args(_,[],[]).
all_args(X,[(_ :- B)|Ps],K) :-
	get_X_vars(X,B,Xv),
	all_args(X,Ps,K1),
	union(Xv,K1,K).

get_X_vars(X,[implies(K,[X])|_],K).
get_X_vars(X,[_|B],K) :-
	get_X_vars(X,B,K).

var_union([],Ys,Ys) :-
	!.
var_union([X|Xs],Ys,Zs) :-
	insert_ord_var(X,Ys,Ys1),
	var_union(Xs,Ys1,Zs).
	
insert_ord_var(X,Ys,Ys) :-
	memb3(X,Ys),
	!.
insert_ord_var(X,Ys,[X|Ys]).

union([],Ys,Ys) :-
	!.
union([X|Xs],Ys,Zs) :-
	insert_ord(X,Ys,Ys1),
	union(Xs,Ys1,Zs).
	
insert_ord(X,L,L) :-
	memb1(X,L),
	!.
insert_ord(X,[],[X]).
insert_ord(X,[Y|Ys],[Y|Zs]) :-
	X @> Y,
	insert_ord(X,Ys,Zs).
insert_ord(X,[Y|Ys],[X,Y|Ys]) :-
	X @< Y.

headclause(P/N,R,(H :- B)) :-
	memb2((H :- B),R),
	functor(H,P,N).
/*
headclause(P/N,R,(H :- B)) :-
	memb2([(H :- B)],R),
	functor(H,P,N).
*/



:- dynamic traceon/0.

writestate(R) :-
	traceon,
	write(R),
	read(_),
	!,
	nl.
writestate(_) :-
	write('*'),
	sp_ttyflush.

display_fix(Cls) :-
        nl,write('Enter filename: '),
        read(F),
        reg_filewrite(F,Cls).

reg_filewrite(tty,Cls) :-
        !,
	nl,write('Prop Abstraction:'),nl,nl,
        displayqTypes(Cls),
	nl,nl.
reg_filewrite(F,Cls) :-
        tell(F),
        displayqTypes(Cls),
        told.

/* write the program to current output file  */

displayqTypes([[]|Ps]) :-
	!,
	displayqTypes(Ps).
displayqTypes([P|Ps]) :-
	displayqRClauses([P]),
	displayqTypes(Ps).
displayqTypes([]).



subatoms(T,Vs) :-
        subatoms(T,[],Vs).

subatoms(X,Vs,Vs1) :-
        atomic(X),
        !,
        insertatom(X,Vs,Vs1).
subatoms(X,Vs,Vs1) :-
        var(X),
        !,
        insertatom(X,Vs,Vs1).
subatoms(X,Vs,Vs1) :-
        nonvar(X),
        X =.. [_|Args],
        argsubatoms(Args,Vs,Vs1).
 
argsubatoms([],Q,Q).
argsubatoms([X|Xs],Vs,Vs2) :-
        subatoms(X,Vs,Vs1),
        argsubatoms(Xs,Vs1,Vs2).
 
insertatom(X,[],[X]).
insertatom(X,[Y|Vs],[Y|Vs]) :-
        X == Y,
        !.
insertatom(X,[Y|Vs],[Y|Vs1]) :-
        insertatom(X,Vs,Vs1).

constant(K) :-
	name(K,[A|_]),
	name(K1,[A]),
	K1 \== 'X'.
