/* computes the determinate unfolding rule */
/* JPG: 10.7.90 */

resultant(A,R) :-
        resultant(A,_,R).

resultant(G,W,R) :-
	copyterm(G,G1),
	canonical(G1),
	argweights(G1,W),
	ref_clause(G,B,_),
	no_Deadend(B),
	prefix(B,R).

prefix(true,true) :-
	!.
prefix(Gs,Gs1) :-
	det_eval(Gs,Gs2),
	!,
	prefix(Gs2,Gs1).
prefix(Gs,Gs).

det_eval(Gs,Gs1) :-
	select_lit(Gs,B,Gs3,N),
	choices(B,Ch,Gs3),
	Ch = [(_,(H :- R))], /* determinate */
	!,
	replace_goal(Gs,N,Gs1,H,R).


select_lit((G,Gs),G,Gs,1).
select_lit((B,Gs),G,(B,Gs1),N1) :-
        !,
        select_lit(Gs,G,Gs1,N),
        N1 is N+1.
select_lit(G,G,true,1).

replace_goal((G,true),2,G,_,_) :- /* eliminate true at end of goal */
	!.
replace_goal((G,Gs),1,Gs1,G,R) :-
	!,
	joingoals(R,Gs,Gs1).
replace_goal((G,Gs),N,(G,Gs1),H,R) :-
	!,
	N1 is N-1,
	replace_goal(Gs,N1,Gs1,H,R).
replace_goal(G,1,R,G,R).

choices(clause(G,B),Cs,Gs) :-
	!,
	bagof((N,(clause(G,B) :- true)),[Gs]^(ref_clause(G,B,N),no_Deadend(B),
		no_Deadend(Gs)),Cs).
choices(G,Ps,Gs) :-
	sp_builtin(G),
	unfoldable(G),
	!,
	pattern(G,P),
	bagof((P,(G :- true)),[Gs]^(call(G),no_Deadend(Gs)), Ps).
choices(G,[],_) :-
	sp_builtin(G),
	!.
choices(G,[],_) :-
	imported(G),
	!.
choices(G,Ps,Gs) :-
	evaluable(G),
	!,
	pattern(G,P),
	bagof((P,(G :- true)),[Gs]^(sp_solve(G),no_Deadend(Gs)), Ps).
choices(G,Cs,Gs) :-
	bagof((N,(G :- B)),
		[Gs]^(ref_clause(G,B,N),no_Deadend(B),no_Deadend(Gs)), Cs).

argweights(T,W) :-
	select_args(T,As),
	!,
	weightlist(As,Ws),
	sumlist(Ws,W).
argweights(T,W) :-
	weight(T,W).

select_args(T,As) :-
	functor(T,P,N),
	functor(T1,P,N),
	ub_clause(measure(T1),true),
	T =.. [P|Xs],
	T1 =.. [P|Ys],
	extract_weighted_args(Xs,Ys,As).

extract_weighted_args([],[],[]).
extract_weighted_args([X|Xs],[1|Ys],[X|Zs]) :-
	extract_weighted_args(Xs,Ys,Zs).
extract_weighted_args([_|Xs],[0|Ys],Zs) :-
	extract_weighted_args(Xs,Ys,Zs).

weight(X,1) :-
	var(X),
	!.
weight(X,N) :-
	X =.. [_|Xs],
	weightlist(Xs,Ws),
	sumlist([1|Ws],N).

weightlist([],[]).
weightlist([X|Xs],[N|Ns]) :-
	weight(X,N),
	weightlist(Xs,Ns).

sumlist([],0).
sumlist([X|Xs],Y) :-
	sumlist(Xs,Y1),
	Y is Y1+X .

equal((A,W),(B,W)) :-
	functor(A,F,N),
	functor(B,F,N).

lessthan((A,[X]),(B,[Y])) :- 
	functor(A,F,N),
	functor(B,F,N),
	X < Y.

leq(X,Y) :-
	equal(X,Y),	
	!.
leq(X,Y) :-
	lessthan(X,Y).
