/* intersect(T1,T2, P):  P is a procedure for a type which is the intersection
   of T1 and T2 */

/* JPG: 16.1.92 */

intersect(any,T2,T2,[],_,N,N) :-
	!.
intersect(T1,any,T1,[],_,N,N) :-
	!.
intersect(T1,T2,T,P,D,N,N) :-
	intersect(T1,T2,T,P,D).
	
intersect(T1,T2,T,[proc(T/1,Def)|Subdefs],D) :-
	sorteddef(T1/1,P1,D),
	sorteddef(T2/1,P2,D),
	intersect_name(T1,T2,T),
	!,
	lower_type(P1,P2,Def,T,pairs([(T1,T2,T)|S],More),D),
%	Def \== [],
	iterate_intersect(More,[(T1,T2,T)|S],Subdefs,D).

iterate_intersect([],_,[],_) :-
	!.
iterate_intersect(M,S,Defs,D) :-
	each_intersect(M,Defs1, M1,S,D),
	iterate_intersect(M1,S,Defs2,D),
	app(Defs1,Defs2,Defs).

each_intersect([],[],_,_,_) :-
	!.
each_intersect([(T1,T2,T)|Ps],[proc(T/1,Def)|Defs],More,S,D) :-
	sorteddef(T1/1,P1,D),
	sorteddef(T2/1,P2,D),
	!,
	lower_type(P1,P2,Def,T,pairs(S,More),D),
	each_intersect(Ps,Defs,More,S,D).

lower_type([],_,[],_,_,_) :-
	!.
lower_type(_,[],[],_,_,_) :-
	!.
lower_type([(H :- B1)|P1],P2,[(S :- B)|Def],T,More,D) :-
	removeclause(H,B2,P2,P3),
	!,
	H =.. [_,K],
	S =.. [T,K],
	intersect_body(B1,B2,B,More,D),
	lower_type(P1,P3,Def,T,More,D).
lower_type([_|P1],P2,Def,T,More,D) :-
	lower_type(P1,P2,Def,T,More,D).

intersect_body(true,true,true,_,_) :-
	!.
intersect_body((B,Bs),(C,Cs),(D,Ds),More,P) :-
	!,
	intersect_atom(B,C,D,More,P),
	intersect_body(Bs,Cs,Ds,More,P).
intersect_body(B,C,D,More,P) :-
	intersect_atom(B,C,D,More,P).


intersect_atom(B,C,C,_,_) :-
	B =.. [any,_],
	!.
intersect_atom(B,C,B,_,_) :-
	C =.. [any,_],
	!.
intersect_atom(B,C,D,More,_) :-
	B =.. [R,X],
	C =.. [S,X],
	get_intersect_name(R,S,T,More),
	D =.. [T,X].


intersect_name(T1,T2,T) :-
	decompose_name(T1,D1),
	decompose_name(T2,D2),
	u_distribute(D1,D2,D3),
	u_normalise(D3,D4),
	compound_is_name(D4,C),
	name(T,C).

u_distribute(u(X,Y),u(Z,W),V) :-
	!,
	i_append(X,Z,A),
	u_distribute(X,W,B),
	u_distribute(Y,Z,C),
	u_distribute(Y,W,D),
	u_append(A,B,E),
	u_append(C,D,F),
	u_append(E,F,V).
u_distribute(u(X,Y),Z,u(V,W)) :-
	!,
	u_distribute(X,Z,V),
	u_distribute(Y,Z,W).
u_distribute(i(X,Y),Z,i(X,W)) :-
	!,
	i_append(Y,Z,W).
u_distribute(X,Y,i(X,Y)).

i_append(i(A,B),C,i(A,D)) :-
	!,
	i_append(B,C,D).
i_append(A,B,i(A,B)).

u_append(u(A,B),C,u(A,D)) :-
	!,
	u_append(B,C,D).
u_append(A,B,u(A,B)).

decompose_name(T,D) :-
	name(T,Tn),
	splitnames(Tn,Ns),
	u_expr(D,Ns,[]).

splitnames([],[]) :-
	!.
splitnames(Ns,[K|Ts]) :-
	gettoken(Ns,T,Ns1),
	name(K,T),
	splitnames(Ns1,Ts).

gettoken([],[],[]).
gettoken([42|Ns],[42],Ns) :- 		% *
	!.
gettoken([43|Ns],[43],Ns) :- 		% +
	!.
gettoken([X,42|Ns],[X],[42|Ns]) :-
	!.
gettoken([X,43|Ns],[X],[43|Ns]) :-
	!.
gettoken([N|Ns],[N|Ms],Ns1) :-
	gettoken(Ns,Ms,Ns1).
	
u_expr(u(X,Y),S,S3) :-
	i_expr(X,S,S1),
	u_symb(S1,S2),
	!,
	u_expr(Y,S2,S3).
u_expr(X,S1,S2) :-
	i_expr(X,S1,S2).

i_expr(i(X,Y),S,S3) :-
	predsymb(X,S,S1),
	i_symb(S1,S2),
	!,
	i_expr(Y,S2,S3).
i_expr(X,S,S1) :-
	predsymb(X,S,S1).

u_symb(['+'|S],S).
i_symb(['*'|S],S).

predsymb(P,[P|S],S) :-
	\+ opsymb(P).

opsymb('*').
opsymb('+').


compound_is_name(u(X,Y),Z) :-
	!,
	compound_is_name(X,X1),	
	compound_is_name(Y,Y1),	
	app(X1,[43|Y1],Z).
compound_is_name(i(X,Y),Z) :-
	!,
	compound_is_name(X,X1),	
	compound_is_name(Y,Y1),	
	app(X1,[42|Y1],Z).
compound_is_name(X,Y) :-
	name(X,Y).


u_normalise(u(X,Y),Z) :-
	!,
	u_sort(X,V),
	u_normalise(Y,V,Z).
u_normalise(X,Z) :-
	u_sort(X,Z).

u_normalise(u(X,Y),V,Z) :-
	!,
	u_sort(X,X1),
	u_insert(X1,V,V1),
	u_normalise(Y,V1,Z).
u_normalise(X,V,Z) :-
	u_sort(X,X1),
	u_insert(X1,V,Z).

u_insert(X,u(Z,Y),u(Z,Y)) :-
	subsumes(Z,X),
	!.
u_insert(X,u(Y,Z),u(Y,W)) :-
	X @> Y,
	!,
	u_insert(X,Z,W).
u_insert(X,u(Y,Z),u(X,u(Y,Z))) :-
	!.
u_insert(X,Z,Z) :-
	subsumes(Z,X),
	!.
u_insert(X,Y,u(Y,X)) :-
	X @> Y,
	!.
u_insert(X,Y,u(X,Y)).

u_sort(i(X,Y),Z) :-
	!,
	u_sort(Y,X,Z).
u_sort(X,X).

u_sort(i(X,Y),Z,W) :-
	!,
	i_insert(X,Z,Z1),
	u_sort(Y,Z1,W).
u_sort(X,Z,W) :-
	i_insert(X,Z,W).

i_insert(X,i(X,Y),i(X,Y)) :-
	!.
i_insert(X,i(Y,Z),i(Y,W)) :-
	X @> Y,
	!,
	i_insert(X,Z,W).
i_insert(X,i(Y,Z),i(X,i(Y,Z))) :-
	!.
i_insert(X,X,X) :-
	!.
i_insert(X,Y,i(Y,X)) :-
	X @> Y,
	!.
i_insert(X,Y,i(X,Y)).

subsumes(X,Y) :-
	\+ i_excess(X,Y).

i_excess(X,Y) :-
	i_memb(Z,X),
	\+ i_memb(Z,Y).

i_memb(X,i(X,_)).
i_memb(X,i(_,Y)) :-
	!,
	i_memb(X,Y).
i_memb(X,X).

get_intersect_name(R,S,T,More) :-
        insert_intersect_pair((S,R,T),More).


insert_intersect_pair(P,pairs(S,M)) :-
        insert_intersect_pair1(P,S,M).

insert_intersect_pair1((A,A,A),S,M) :-
        var(S),
        !,
        S = [(A,A,A)|_],
        memb1((A,A,A),M).
insert_intersect_pair1((A,B,C),S,M) :-
        var(S),
        !,
        intersect_name(A,B,C),
        S = [(A,B,C)|_],
        memb1((A,B,C),M).
insert_intersect_pair1(P,[P|_],_) :-
        !.
insert_intersect_pair1(P,[_|S],M) :-
        insert_intersect_pair1(P,S,M).

sorteddef(T,D,[proc(T,D)|_]) :-
	!.
sorteddef(T,D,[_|Ds]) :-
	sorteddef(T,D,Ds).
