m(F,F2) :-
	represent(F,Cls),
	modelize(Cls,Ms,L,BI),
	mylist(L),
	mylist(BI),
	generate_int(L,Int),
	app(Int,Ms,Bigprog),
	tell(F2),
	write_reduced_clauses(Bigprog),
	told,
	write('The following builtins occurred in the program: '),
	nl,
	write(BI).

modelize([Cl1|Cls], [(H :- R)|Rs],L,BI) :-
%modelize([Cl|Cls], [(H :- R)|Rs],L,BI) :-
	exec_eq(Cl1,Cl),
	domains(Cl,Ds,Pairs),
	flatten_args(Cl,(H :- B),Pairs,L),
	removetrue(Ds,Ds1),
	joingoals(B,Ds1,R),
	look_for_builtins(R,BI),
	modelize(Cls,Rs,L,BI).
modelize([],[],_,_).

look_for_builtins(true,_) :-
	!.
look_for_builtins((B,Bs),BI) :-
	sp_builtin(B),
	functor(B,F,N),
	!,
	memb1(F/N,BI),
	look_for_builtins(Bs,BI).
look_for_builtins((_,Bs),BI) :-
	!,
	look_for_builtins(Bs,BI).
look_for_builtins(B,BI) :-
	sp_builtin(B),
	functor(B,F,N),
	!,
	memb1(F/N,BI).
look_for_builtins(_,_).


exec_eq((H :- B),(H :- B1)) :- 
	!,
	exec_eq(B,B1).
exec_eq(true,true) :-
	!.
exec_eq((X=X,Bs),Bs1) :-
	!,
	exec_eq(Bs,Bs1).
exec_eq((B,Bs),(B,Bs1)) :-
	!,
	exec_eq(Bs,Bs1).
exec_eq(X=X,true) :-
	!.
exec_eq(B,B).


nested(X,(Den,_)) :-
	vars(T,Vs),
	flatten_denotes(denotes(T,_),Den),
	memb3(X,Vs).
nested(X,(_,Bs)) :-
	nested(X,Bs).

new_p(P,N) :-
	name(p,Pn),
	name(N,Nn),
	app(Pn,[95|Nn],PNn),
	name(P,PNn).

flatten_args(X,Y,(Den,Ps),L) :-
	inner_clause_subterm(X,T,X1,U),
	!,
	flatten_denotes(denotes(T,U),Den),
	functor(T,F,N),
	memb1(F/N,L),
	flatten_args(X1,Y,Ps,L).
flatten_args(X,X,true,_).

domains(_,D,D).
/*
domains((H :- B),D,D1) :-
	vars(H,Xs),
	vars(B,Ys),
	domainvars(Xs,Ys,D,D1).
*/
	
domainvars([],_,D,D).
domainvars([X|Xs],Ys,D,D1) :-
	memb3(X,Ys),
	!,
	domainvars(Xs,Ys,D,D1).
domainvars([X|Xs],Ys,(domain(X),D),D1) :-
	domainvars(Xs,Ys,D,D1).

% subterm(T,S,T1,S1):  S is a subterm of T;  
% S1 is a variable that occurs in T1 at the place where S was in T.

inner_clause_subterm((H :- B),T,(X1 :- B),T1) :-
	inner_atom_subterm(H,T,X1,T1).
inner_clause_subterm((H :- B),T,(H :- X1),T1) :-
	inner_body_subterm(B,T,X1,T1).

inner_atom_subterm(H,T,H1,T1) :-
	H =.. [P|Xs],
	inner_arg_subterm(Xs,T,Xs1,T1),
	H1 =.. [P|Xs1].

inner_body_subterm((B,Bs),T,(X,Bs),T1) :-
	inner_atom_subterm(B,T,X,T1).
inner_body_subterm((B,Bs),T,(B,X),T1) :-
	!,
	inner_body_subterm(Bs,T,X,T1).
inner_body_subterm(B,T,X,T1) :-
	inner_atom_subterm(B,T,X,T1).

inner_arg_subterm([X|Xs],T,[X1|Xs],T1) :-
	inner_subterm(X,T,X1,T1).
inner_arg_subterm([X|Xs],T,[X|Xs1],T1) :-
	inner_arg_subterm(Xs,T,Xs1,T1).

inner_subterm(X,T,X1,T1) :-
	subterm(X,T,X1,T1),
	nonvar(T),
	T =.. [_|Zs],
	allvars(Zs).

allvars([]).
allvars([X|Xs]) :-
	var(X),
	allvars(Xs).

subterm(T,T,X,X) :-
	nonvar(T).
subterm(T,R,X,Y) :-
	nonvar(T),
	T =.. [F|Xs],		
	replace(Xs,R1,Ys,Y1),
	subterm(R1,R,Y1,Y),
	X =.. [F|Ys].
	
replace([X|Xs],X,[Y|Xs],Y).
replace([X|Xs],U,[X|Ys],Y) :-
	replace(Xs,U,Ys,Y).

represent(F,Cs) :-
        see(F),
        read(C),
        rep_clauses(C,Cs,N),
        seen,
	write(N),
	write(' clauses were read'),
	nl,
	write('Finished reading '),
	write(F),
	nl,
        !.
represent(F,[],0) :-
        seen,
        write('Cannot find file '), write(F),
        write(', or problems while reading.'), nl,nl.

rep_clauses(C,[],0) :-
        end_of_file(C),
        !.
rep_clauses((:- op(X,Y,Z)),Cs,N) :-
        !,
        call(op(X,Y,Z)),
        read(C1),
        rep_clauses(C1,Cs,N).
rep_clauses((:- _),Cs,N) :-
        !,
        read(C1),
        rep_clauses(C1,Cs,N).
rep_clauses((H :- B),[(H :- B)|Cs],N) :-
	!,
        read(C1),
        rep_clauses(C1,Cs,N1),
	N is N1+1.
rep_clauses(H,[(H :- true)|Cs],N) :-
        read(C1),
        rep_clauses(C1,Cs,N1),
	N is N1+1.

%eof(end_of_file).




write_reduced_clauses([]).
write_reduced_clauses([C|Cs]) :-
	%canonical(C),
	writeq(C),
	%sp_writeClause(C),
	write('.'),
	nl,
	write_reduced_clauses(Cs).

removetrue((B,true),B) :-
	!.
removetrue(true,true).
removetrue((B,Bs),(B,Bs1)) :-
	removetrue(Bs,Bs1).

mylist([]) :-
	!.
mylist([_|X]) :-
	mylist(X).

generate_int(L,Int) :-
	write('generate interpretation? (y/n) '),
	getchar(YN),
	gen_int(YN,L,Int).

gen_int(n,_,[]).
%gen_int(y,L,Int) :-
%	getdomain(D,Dlist),
%	gen_each_symb(L,Dlist,Den),
%	app(D,Den,Int).
gen_int(y,L,Den) :-
	getdomain(Dlist),
	gen_each_symb(L,Dlist,Den).

%getdomain(D,Dlist) :-
%	write('Domain (finite list): '),
%	read(Dlist),
%	domainfacts(Dlist,D).
getdomain(Dlist) :-
	write('Domain (finite list): '),
	read(Dlist).

domainfacts([],[]).
domainfacts([D|Ds],[domain(D)|DFs]) :-
	domainfacts(Ds,DFs).

gen_each_symb([F/N|L],D,Int) :-
	cart_product(D,N,DN),
	askuser(DN,F,Int, Restint),
	gen_each_symb(L,D,Restint).
gen_each_symb([],_,[]).

cart_product(_,0,[[]]).
cart_product(D,N,DN) :-
	N > 0,
	N1 is N-1,
	cart_product(D,N1,DN1),
	mult_d(D,DN1,DN).

mult_d(_,[],[]).
mult_d(D,[T|Ts],DN) :-
	mult_tuple(D,T,DN,Q),
	mult_d(D,Ts,Q).

mult_tuple([],_,Q,Q).
mult_tuple([X|Xs],T,[[X|T]|Ts],Q) :-
	mult_tuple(Xs,T,Ts,Q).

askuser([],_,I,I).
askuser([T|Ts],F,[Den|Int], Restint) :-
	A =.. [F|T],
	write(A),
	write(' = '),
	read(V),
	flatten_denotes(denotes(A,V),Den),
	askuser(Ts,F,Int, Restint).

flatten_denotes(denotes(T,V),Den) :-
	T =.. [F|Xs],
	name(denotes,Dn),
	name(F,Fn),
	app(Dn,[95|Fn],DFn),
	name(DF,DFn),
	app(Xs,[V],Ys),
	Den =.. [DF|Ys].
