bu(F) :-
	represent(F,Cs),
	time(freduce(Cs,Q2)),
	tell(temp),	
	write_reduced_clauses(Q2),
	told,
	sp_consult(temp),
        program_preds(P),
        tp_fix(P).

td(F,Q) :-
	time((magic(F,temp1,Q),
	represent(temp1,Cs),
	freduce(Cs,Q2))),
	tell(temp),	
	write_reduced_clauses(Q2),
	told,
	sp_consult(temp),
	program_preds(P),
	tp_fix(P).

freduce(Cs,Q2) :-
	type_reduce_clauses(Cs,Q),
	body_transform(Q,Q1),
	head_transform(Q1,Q2,1).

type_reduce_clauses([Cl|Cls], [R|Rs]) :-
	reduce(Cl,R),
	canonical(R),
	type_reduce_clauses(Cls,Rs).
type_reduce_clauses([], []).

reduce(T,R) :-
	transform(T,T1),
	!,
	reduce(T1,R).
reduce(T,T).

transform(T,T1) :-
	subterm(T,S,T1,S1),
	findall((S,S1),rule(S,S1),[(S,S1)]).

% subterm(T,S,T1,S1):  S is a subterm of T;  
% S1 is a variable that occurs in T1 at the place where S was in T.

subterm(T,T,X,X) :-
	nonvar(T).
subterm(T,R,X,Y) :-
	nonvar(T),
	T =.. [F|Xs],		
	replace(Xs,R1,Ys,Y1),
	subterm(R1,R,Y1,Y),
	X =.. [F|Ys].
	
replace([X|Xs],X,[Y|Xs],Y).
replace([X|Xs],U,[X|Ys],Y) :-
	replace(Xs,U,Ys,Y).

represent(F,Cs) :-
        see(F),
        read(C),
        rep_clauses(C,Cs,N),
        seen,
	write(N),
	write(' clauses were read'),
	nl,
	write('Finished reading '),
	write(F),
	nl,
        !.
represent(F,[],0) :-
        seen,
        write('Cannot find file '), write(F),
        write(', or problems while reading.'), nl,nl.

rep_clauses(C,[],0) :-
        eof(C),
        !.
rep_clauses((:- op(X,Y,Z)),Cs,N) :-
        !,
        call(op(X,Y,Z)),
        read(C1),
        rep_clauses(C1,Cs,N).
rep_clauses((:- _),Cs,N) :-
        !,
        read(C1),
        rep_clauses(C1,Cs,N).
rep_clauses((H :- B),[(H :- B)|Cs],N) :-
	!,
        read(C1),
        rep_clauses(C1,Cs,N1),
	N is N1+1.
rep_clauses(H,[(H :- true)|Cs],N) :-
        read(C1),
        rep_clauses(C1,Cs,N1),
	N is N1+1.

eof(end_of_file).


rule([X0|list(X1)], list(union(X0,X1))).
rule([], list(bottom)).
%rule(empty,tree(bottom)).
%rule(tree(X0,tree(X1),tree(X2)),tree(union(X0,union(X1,X2)))).
rule(X,num) :-
	nonvar(X),
	number(X).
rule(num+num, num).


head_transform([],[],_).
head_transform([C|Cs],Cs1,K) :-
	clause_head_transform(C,C1,K),
	!,
	K1 is K+1,
	head_transform([C1|Cs],Cs1,K1).
head_transform([C|Cs],[C|Cs1],K) :-
	head_transform(Cs,Cs1,K).

clause_head_transform((H :- B),(H1 :- (union(X,Y,Z),B)),K) :-
	subterm(H,_,union(X,Y),H1,Z),
	\+ unionsubterm(X),
	\+ unionsubterm(Y),
	newvar(K,Z).

unionsubterm(X) :-
	subterm(X,_,union(_,_),_,_).

body_transform([],[]).
body_transform([C|Cs],[C1|Cs1]) :-
	clause_body_transform(C,C1),
	body_transform(Cs,Cs1).

clause_body_transform((H :- B), (H :- B1)) :-
	body_lits_transform(B,B1).

body_lits_transform((B,Bs),Bs1) :-
	subterm(B,Path,union(X,Y),B1,X),
	!,
	subterm(B,Path,union(X,Y),B2,Y),
	body_lits_transform((B1,B2,Bs),Bs1).
body_lits_transform((B,Bs),(B,Bs1)) :-
	!,
	body_lits_transform(Bs,Bs1).
body_lits_transform(B,Bs1) :-
	subterm(B,Path,union(X,Y),B1,X),
	!,
	subterm(B,Path,union(X,Y),B2,Y),
	body_lits_transform((B1,B2),Bs1).
body_lits_transform(B,B).


subterm(T,[],T,X,X).
subterm(T,[N|Ns],R,X,Y) :-
	T =.. [F|Xs],		
	replace(Xs,N,R1,Ys,Y1),
	subterm(R1,Ns,R,Y1,Y),
	X =.. [F|Ys].
	
replace([X|Xs],1,X,[Y|Xs],Y).
replace([X|Xs],N,U,[X|Ys],Y) :-
	replace(Xs,N1,U,Ys,Y),
	N is N1+1.

write_reduced_clauses([]).
write_reduced_clauses([C|Cs]) :-
	write(C),
	write('.'),
	nl,
	write_reduced_clauses(Cs).

newvar(K,P) :-
        name(K,Kn),
        name('Z',T),
        app(T,Kn,Tn),
        name(P,Tn).

newlub(K,P) :-
        name(K,Kn),
        name(union,T),
        app(T,Kn,Tn),
        name(P,Tn).
