r2c(F1,F2) :-
	represent(F1,P),
	defined_funcs(P,Fs),
	mylist(Fs),
	transform_rules(P,Fs,Q),
	tell(F2),
	write_rule_clauses(Q),
	told.

defined_funcs([], _) :-
	!.
defined_funcs([(rule(L,_) :- true)|Rs],Fs) :-
	!,
	functor(L,F,N),
	memb1(F/N,Fs),
	defined_funcs(Rs,Fs).
defined_funcs([_|Rs],Fs) :-
	defined_funcs(Rs,Fs).

transform_rules([(rule(L,R) :- true)|Rs],Fs,[(L1 :- R1)|Q]) :-
	!,
	rule2clause(L,R,L1,R1,Fs),	
	transform_rules(Rs,Fs,Q).
transform_rules([_|Rs],Fs,Q) :-
	transform_rules(Rs,Fs,Q).
transform_rules([],_,[]).
	
rule2clause(L,R,L1,R1,Fs) :-
%	L =.. [P|Xs],
	rule_clause_body(L,LB,Fs,Z),
	head_atom(LB,L1,LB1),
	rule_clause_body(R,R2,Fs,Z),
	joingoals(R2,LB1,R3),
	removetrue(R3,R1).

head_atom((L1,true),L1,true).
head_atom((A,B,Bs),L1,(A,Bs1)) :-
	head_atom((B,Bs),L1,Bs1).

rule_clause_body(R,(B,Bs),Fs,Z) :-
	subterm(R,S,R1,U),
	functor(S,G,M),
	memb1(G/M,Fs),
	innermost(S,Fs),
	!,
	S =.. [P|Xs],
	app(Xs,[U],Ys),
	new_pred(P,P1),
	B =.. [P1|Ys],
	rule_clause_body(R1,Bs,Fs,Z).
rule_clause_body(R,true,_,R) :-
	var(R),
	!.
rule_clause_body(R,R,Fs,Z) :-
	functor(R,G,M),
	old_pred(G,G1),
	M1 is M-1,
	memb1(G1/M1,Fs),
	!,
	arg(M,R,Z).
rule_clause_body(R,true,_,R).
	
	
innermost(S,_) :-
	var(S),
	!.
innermost(S,Fs) :-
	S =.. [_|Xs],
	\+ contains_defined_func(Xs,Fs).

contains_defined_func([R|_],Fs) :-
	subterm(R,S,_,_),
	functor(S,G,M),
	memb1(G/M,Fs).
contains_defined_func([_|Rs],Fs) :-
	contains_defined_func(Rs,Fs).
	
new_pred(P,P1) :-
	name(P,Pn),
	name('_rel',Suffix),
	app(Pn,Suffix,PNn),
	name(P1,PNn).

old_pred(P,P1) :-
	name(P,Pn),
	name('_rel',Suffix),
	app(PNn,Suffix,Pn),
	name(P1,PNn).


write_rule_clauses([]).
write_rule_clauses([C|Cs]) :-
	canonical(C),
	write(C),
	write('.'),
	nl,
	write_rule_clauses(Cs).

