/* subtype(T1,T2,D):  T1 is a sub-type of T2 with set of type defs D */
/* T1 and T2 are atoms whose procedures are
   defined by Regular Unary Logic Programs */

/* JPG:  15.1.92   */

%:- setdebug.
% :- dynamic user_proc/3.

eqtype(T1,T2,D) :-
	subtype(T1,T2,D),
	subtype(T2,T1,D).

subtype(T1,T2,D) :-
	subtype(T1,T2,[],_,D).

subtype(_,freevar,S,S,_) :-
	!.
subtype(_,sharevar,S,S,_) :-
	!.
subtype(T,T,S,S,_) :-
	!.
subtype(T1,T2,S,S,_) :-
	loopsubtype(T1,T2,S),
	!.
subtype(T1,T2,S,S1,D) :-
	T1 \== freevar,
	T1 \== sharevar,
	T2 \== freevar,
	T2 \== sharevar,
	def(T1/1,P1,D),
	def(T2/1,P2,D),
	!,
	subtype1(P1,P2,[(T1,T2)|S],S1,D).


subtype1([],_,S,S,_).
subtype1([(H :- B)|P1],P2,S,S2,D) :-
	removeclause(H,B1,P2,P3),
	subtype_pairs(B,B1,S,S1,D),
	subtype1(P1,P3,S1,S2,D).
	
subtype_pairs(true,true,S,S,_) :-
	!.
subtype_pairs((B,Bs),(D,Ds),S,S2,P) :-
	!,
	subtype2(B,D,S,S1,P),
	subtype_pairs(Bs,Ds,S1,S2,P).
subtype_pairs(B,D,S,S1,P) :-
	subtype2(B,D,S,S1,P).

subtype2(B,D,S,S1,P) :-
	functor(B,T1,1),
	functor(D,T2,1),
	subtype(T1,T2,S,S1,P).

removeclause(H,B,[(G :- B)|P],P) :-
	H =.. [_,T1],
	G =.. [_,T2],
	functor(T1,F,N),
	functor(T2,F,N),
	!.
removeclause(H,B,[X|P],[X|P1]) :-
	removeclause(H,B,P,P1).

loopsubtype(X,Y,[(X,Y)|_]) :-
	!.
loopsubtype(X,Y,[_|S]) :-
	loopsubtype(X,Y,S).

user_clauses(P/N,Cls) :-
        functor(H,P,N),
        findall((H:-B),ub_clause(H,B),Cls),
	!.
user_clauses(_,[]).

% user_clauses(P/N,Cls) :-
% 	user_proc(P,N,Cls).

	
def(T,P,D) :-
	def1(T,P,D),
	!.
def(T,P,_) :-
	user_clauses(T,P),
	canonical(P),
	P \== [],
	!.
def(T,_,_) :-
	write(T),write(' not found'),nl,fail.

def1(T,P,[[]|Ps]) :-
	def1(T,P,Ps),
	!.
def1(T,P,[[Q|Qs]|_]) :-
	def1(T,P,[Q|Qs]),
	!.
def1(T,P,[[_|_]|Ps]) :-
	def1(T,P,Ps),
	!.

def1(T/1,P,[proc(T/1,P)|_]) :-
	!.
def1(T,P,[proc(_,_)|D]) :-
	def1(T,P,D).

	
