/* ub(T1,T2,T3,P,R,K,K1):  P is a procedure for a type which is an upper
   bound of T1 and T2 */

/* JPG: 16.1.92 */

%ub(T1,T2,T2,[],D,N,N) :-
%	subtype(T1,T2,D),
%	!.
%ub(T1,T2,T1,[],D,N,N) :-
%	subtype(T2,T1,D),
%	!.
%ub(T1,T2,T3,R,D,N,N1) :-
%	ub1(T1,T2,T3,R,D,N,N1).
%
ub1(T1,T2,T,[proc(T/1,Def)|Subdefs],D,N,N3) :-
	def(T1/1,P1,D),
	def(T2/1,P2,D),
	newname(N,T),
	N1 is N+1,
	!,
	upper_type(P1,P2,Def,T,pairs([(T1,T2,T)|S],More),N1,N2,D),
	iterate_ub(More,[(T1,T2,T)|S],Subdefs,N2,N3,D).

iterate_ub([],_,[],N,N,_) :-
	!.
iterate_ub(M,S,Defs,N,N2,D) :-
	each_ub(M,Defs1, M1,S,N,N1,D),
	iterate_ub(M1,S,Defs2,N1,N2,D),
	app(Defs1,Defs2,Defs).

each_ub([],[],_,_,N,N,_) :-
	!.
each_ub([(T1,T2,T)|Ps],[proc(T/1,Def)|Defs],More,S,N,N3,D) :-
	def(T1/1,P1,D),
	def(T2/1,P2,D),
	N1 is N+1,
	!,
	upper_type(P1,P2,Def,T,pairs(S,More),N1,N2,D),
	each_ub(Ps,Defs,More,S,N2,N3,D).

upper_type([],[],[],_,_,N,N,_) :-
	!.
upper_type([],[(H :- B)|P2],[(S :- B)|Def],T,More,N,N1,D) :-
	!,
	H =.. [_,K],
	S =.. [T,K],
	upper_type([],P2,Def,T,More,N,N1,D).
upper_type([(H :- B)|P1],[],[(S :- B)|Def],T,More,N,N1,D) :-
	!,
	H =.. [_,K],
	S =.. [T,K],
	upper_type(P1,[],Def,T,More,N,N1,D).
upper_type([(H :- true)|P1],P2,[(S :- true)|Def],T,More,N,N2,D) :-
	H =.. [_,K],
	basic_type(K),
	!,
	remove_type_instances(K,P2,P3),
	S =.. [T,K],
	upper_type(P1,P3,Def,T,More,N,N2,D).
upper_type([(H :- true)|P1],P2,[(S :- true)|Def],T,More,N,N2,D) :-
	H =.. [_,K],
	basic_type_instance(K,P2,K1,P3),
	!,
	remove_type_instances(K1,P1,P4),
	S =.. [T,K1],
	upper_type(P4,P3,Def,T,More,N,N2,D).
upper_type([(H :- B1)|P1],P2,[(S :- B)|Def],T,More,N,N2,D) :-
	removeclause1(H,B2,P2,P3),
	!,
	H =.. [_,K],
	S =.. [T,K],
	ub_body(B1,B2,B,More,N,N1,D),
	upper_type(P1,P3,Def,T,More,N1,N2,D).
upper_type([(H :- B1)|P1],P2,[(S :- B1)|Def],T,More,N,N2,D) :-
	H =.. [_,K],
	S =.. [T,K],
	upper_type(P1,P2,Def,T,More,N,N2,D).

basic_type_instance(K,[(H :- true)|P],K1,P) :-
	H =.. [_,K1],
	basic_type(K1),
	type_check(K,K1),
	!.
basic_type_instance(K,[C|P],K1,[C|P1]) :-
	basic_type_instance(K,P,K1,P1).

remove_type_instances(_,[],[]).
remove_type_instances(K,[(H :- true)|P1],P2) :-
	H =.. [_,A],
	type_check(A,K),
	!,
	remove_type_instances(K,P1,P2).
remove_type_instances(K,[(H :- true)|P1],P2) :-
	H =.. [_,K],
	!,
	remove_type_instances(K,P1,P2).
remove_type_instances(K,[Cl|P1],[Cl|P2]) :-
	remove_type_instances(K,P1,P2).
	
removeclause1(H,B,[(G :- B)|P],P) :-
        H =.. [_,T1],
        G =.. [_,T2],
        functor(T1,F,N),
        functor(T2,F,N),
        !.
removeclause1(H,B,[X|P],[X|P1]) :-
        removeclause1(H,B,P,P1).

ub_body(true,true,true,_,N,N,_) :-
	!.
ub_body((B,Bs),(C,Cs),(D,Ds),More,N,N2,P) :-
	!,
	ub_atom(B,C,D,More,N,N1,P),
	ub_body(Bs,Cs,Ds,More,N1,N2,P).
ub_body(B,C,D,More,N,N1,P) :-
	ub_atom(B,C,D,More,N,N1,P).

/*

ub_atom(B,C,C,_,N,N,D) :-
	functor(B,R,1),
	functor(C,S,1),
	subtype(R,S,D),
	!.
ub_atom(B,C,B,_,N,N,D) :-
	functor(B,R,1),
	functor(C,S,1),
	subtype(S,R,D),
	!.
%ub_atom(_,any(X),any(X),_,N,N,_) :-
%	!.
%ub_atom(any(X),_,any(X),_,N,N,_) :-
%	!.
ub_atom(B,C,D,More,N,N1,_) :-
	B =.. [R,X],
	C =.. [S,X],
	get_name(R,S,T,More,N,N1),
	D =.. [T,X].

*/

ub_atom(B,C,D,More,N,N1,P) :-
        B =.. [R,X],
        C =.. [S,X],
        next_ub(R,S,T,More,N,N1,P),
        D =.. [T,X].

next_ub(R,S,S,_,N,N,D) :-
        subtype(R,S,D),
        !.
next_ub(R,S,R,_,N,N,D) :-
        subtype(S,R,D),
        !.
next_ub(R,S,T,More,N,N1,_) :-
        get_name(R,S,T,More,N,N1).


get_name(R,S,T,More,N,N1) :-
	R @=< S,
	!,
	insert_pair((R,S,T),More,N,N1).
get_name(R,S,T,More,N,N1) :-
	insert_pair((S,R,T),More,N,N1).
	

insert_pair(P,pairs(S,M),N,N1) :-
	insert_pair1(P,S,M,N,N1).

insert_pair1((A,B,C),S,M,N,N1) :-
	var(S),
	!,
	newname(N,C),
	N1 is N+1,
	S = [(A,B,C)|_],
	memb1((A,B,C),M).
insert_pair1(P,[P|_],_,N,N) :-
	!.
insert_pair1(P,[_|S],M,N,N1) :-
	insert_pair1(P,S,M,N,N1).

/* normalise(Def1, T2, T,Def2):  Def1 is the type definition for
   type T, and every occurrence of T2
   in the bodies of Def1 are replaced by T.  */


normalise_defs([],_,_,[]).
normalise_defs([proc(F,Cls)|Ps],T2,T,[proc(F,Cls1)|Ps1]) :-
        normalise(Cls,T2,T,Cls1),
        normalise_defs(Ps,T2,T,Ps1).


replaceproc(T1,[proc(T1/1,_)|Ps],Cls,[proc(T1/1,Cls)|Ps]) :-
        !.
replaceproc(T1,[P|Ps],Cls,[P|Ps1]) :-
        replaceproc(T1,Ps,Cls,Ps1).

normalise([], _,_,[]).
normalise([(H :- B)|Cls], T2,F,[(H :- B1)|T]) :-
        normalise_body(B,T2,F,B1),
        normalise(Cls,T2,F,T).

normalise_body(true,_,_,true) :-
        !.
normalise_body((B,Bs),T2,F,(B1,Bs1)) :-
        B =.. [T2,X],
        !,
        B1 =.. [F,X],
        normalise_body(Bs,T2,F,Bs1).
normalise_body((B,Bs),T2,F,(B,Bs1)) :-
        !,
        normalise_body(Bs,T2,F,Bs1).
normalise_body(B,T2,F,B1) :-
        B =.. [T2,X],
        !,
        B1 =.. [F,X].
normalise_body(B,_,_,B).


/* normalise(Def1, T1,T2, T3,Def2):  replace all occurrences
   of T1 and T2 in Def1 by T3, yielding Def2 */


/*
normalise_defs([],_,_,_,[]).
normalise_defs([proc(F,Cls)|Ps],T1,T2,T3,[proc(F1,Cls1)|Ps1]) :-
	normalise_proc_id(T1,T2,T3,F,F1),
	normalise(Cls,T1,T2,T3,Cls1),
	normalise_defs(Ps,T1,T2,T3,Ps1).


replaceproc(T1,[proc(T1/1,_)|Ps],Cls,[proc(T1/1,Cls)|Ps]) :-
	!.
replaceproc(T1,[P|Ps],Cls,[P|Ps1]) :-
	replaceproc(T1,Ps,Cls,Ps1).

normalise([],_,_,_,[]).
normalise([(H :- B)|Cls], T1,T2,T3,[(H1 :- B1)|T]) :-
	normalise_head(T1,T2,T3,H,H1),
	normalise_body(B,T1,T2,T3,B1),
	normalise(Cls,T1,T2,T3,T).

normalise_proc_id(T1,_,T3,T1/1,T3/1) :-
	!.
normalise_proc_id(_,T2,T3,T2/1,T3/1) :-
	!.
normalise_proc_id(_,_,_,H,H).

normalise_head(T1,_,T3,H,H1) :-
	H =.. [T1|X],
	H1 =.. [T3|X],
	!.
normalise_head(_,T2,T3,H,H1) :-
	H =.. [T2|X],
	H1 =.. [T3|X],
	!.
normalise_head(_,_,_,H,H).

normalise_body(true,_,_,_,true) :-
	!.
normalise_body((B,Bs),T1,T2,T3,(B1,Bs1)) :-
	B =.. [T2,X],
	!,
	B1 =.. [T3,X],
	normalise_body(Bs,T1,T2,T3,Bs1).
normalise_body((B,Bs),T1,T2,T3,(B1,Bs1)) :-
	B =.. [T1,X],
	!,
	B1 =.. [T3,X],
	normalise_body(Bs,T1,T2,T3,Bs1).
normalise_body((B,Bs),T1,T2,T3,(B,Bs1)) :-
	!,
	normalise_body(Bs,T1,T2,T3,Bs1).
normalise_body(B,_,T2,T3,B1) :-
	B =.. [T2,X],
	!,
	B1 =.. [T3,X].
normalise_body(B,T1,_,T3,B1) :-
	B =.. [T1,X],
	!,
	B1 =.. [T3,X].
normalise_body(B,_,_,_,B).
*/
	
/* get all clauses on which a predicate depends */	

rdefs(T,Defs,P) :-
	iterate_defs([T],[T|_],Defs,[],P).

iterate_defs([],_,D,D,_) :-
	!.
iterate_defs(Ts,Ss,Defs,Defs2,P) :-
	each_def(Ts,Defs,Defs1,Ss,More,P),
	iterate_defs(More,Ss,Defs1,Defs2,P).

each_def([],D,D,_,_,_) :-
	!.
each_def([T|Ts],[proc(T/1,P)|Ps],Defs,Ss,M,D) :-
	def(T/1,P,D),
	body_defs(P,Ss,M,D),
	each_def(Ts,Ps,Defs,Ss,M,D).

body_defs([(_ :- B)|Cs],Ss,M,P) :-
	bodytypes(B,Ss,M,P),
	body_defs(Cs,Ss,M,P).
body_defs([],_,_,_).

bodytypes((B,Bs),Ss,M,P) :-
	!,
	functor(B,T,1),
	insertT(T,Ss,M),
	bodytypes(Bs,Ss,M,P).
bodytypes(true,_,_,_) :-
	!.
bodytypes(B,Ss,M,_) :-
	functor(B,T,1),
	insertT(T,Ss,M).

insertT(any,_,_) :-
	!.
insertT(T,S,M) :-
	var(S),
	!,
	memb1(T,M),
	S = [T|_].
insertT(T,[T|_],_) :-
	!.
insertT(T,[_|S],M) :-
	insertT(T,S,M).


rename_all_preds([],[],_,K,K).
rename_all_preds([proc(T/1,Cls)|Ps],[proc(T1/1,Cls1)|Ps1],S,K0,K3) :-
	renamePred(T,T1,S,K0,K1),
	rename_clauses(T1,Cls,Cls1,S,K1,K2),
	rename_all_preds(Ps,Ps1,S,K2,K3).

renamePred(T,T1,S,K0,K1) :-
	var(S),
	!,
	newname(K0,T1),
	K1 is K0+1,
	S = [(T,T1)|_].
renamePred(T,T1,[(T0,T1)|_],K,K) :-
	T == T0,
	!.
renamePred(T,T1,[_|S],K0,K1) :-
	renamePred(T,T1,S,K0,K1).

rename_clauses(_,[],[],_,K,K).
rename_clauses(T1,[(H :- B)|Cls],[(H1 :- B1)|Cls1],S,K0,K2) :-
	H =.. [_|Y],
	H1 =.. [T1|Y],
	rename_body(B,B1,S,K0,K1),
	rename_clauses(T1,Cls,Cls1,S,K1,K2).
	
rename_body(true,true,_,K,K) :-
	!.
rename_body((B,Bs),(B,Bs1),S,K0,K1) :-
	B =.. [any,_],
	!,
	rename_body(Bs,Bs1,S,K0,K1).
rename_body((B,Bs),(B,Bs1),S,K0,K1) :-
	B =.. [T,_],
	basic_type(T),
	!,
	rename_body(Bs,Bs1,S,K0,K1).
rename_body((B,Bs),(B1,Bs1),S,K0,K2) :-
	!,
	B =.. [T,X],
	renamePred(T,T1,S,K0,K1),
	B1 =.. [T1,X],
	rename_body(Bs,Bs1,S,K1,K2).
rename_body(B,B,_,K,K) :-
	B =.. [any,_],
	!.
rename_body(B,B,_,K,K) :-
	B =.. [T,_],
	basic_type(T),
	!.
rename_body(B,B1,S,K0,K1) :-
	B =.. [T,X],
	renamePred(T,T1,S,K0,K1),
	B1 =.. [T1,X].
