/* regularise(P,D):  D is set of RUL clauses which 
   is not regular program.  P is RUL program. */

/* JPG: 16.1.92 */ 
/* JPG: 13.8.93 adapted from ub.pl */ 


regularise(P,_,R,D1,D2) :-
	iterate_ub_merge([(P,P,P)|_],[(P,P,P)|_],R1,R2,D1,D2),
	app(R1,R2,R).

iterate_ub_merge([],_,D1,D2,D1,D2) :-
	!.
iterate_ub_merge(M,S,R1,R2,D1,D2) :-
	each_ub_merge(M,M1,S,D1,D2,D3,D4),
	iterate_ub_merge(M1,S,R1,R2,D3,D4).

each_ub_merge([],_,_,D1,D2,D1,D2) :-
	!.
each_ub_merge([(T,T,T)|Ps],More,S,D1,D2,R1,R2) :-
	!,
	rul_defs(T/1,P1,D1,D2,D3,D4),
	merge_rul(P1,Def,pairs(S,More)),
	each_ub_merge(Ps,More,S,D3,[proc(T/1,Def)|D4],R1,R2).
each_ub_merge([(T1,T2,T3)|Ps],More,S,D1,D2,R1,R2) :-
	!,
	rul_defs(T1/1,P1,D1,D2),
	rul_defs(T2/1,P2,D1,D2),
	app(P1,P2,P3),
	renametypes(P3,T3,P4),
	merge_rul(P4,Def,pairs(S,More)),
	rul_defs(T3/1,_,D1,D2,D3,D4),
	each_ub_merge(Ps,More,S,D3,[proc(T3/1,Def)|D4],R1,R2).

merge_rul(P,Q,More) :-
	remove((H :- B1),P,P1),
	remove_similar(H,B1,B2,P1,P2),	
	!,
	ub_body_merge(B1,B2,B,More),
	merge_rul([(H :- B)|P2],Q,More).
merge_rul(P,P,_).

remove_similar(H,B1,B1,P1,P2) :-	
	removeclause(H,B1,P1,P2).	% two identical clauses 
remove_similar(H,_,B2,P1,P2) :-
	removeclause(H,B2,P1,P2).	% same head only


ub_body_merge(true,true,true,_) :-
	!.
ub_body_merge((B,Bs),(C,Cs),(D,Ds),More) :-
	!,
	ub_atom_merge(B,C,D,More),
	ub_body_merge(Bs,Cs,Ds,More).
ub_body_merge(B,C,D,More) :-
	ub_atom_merge(B,C,D,More).

ub_atom_merge(_,C,C,_) :-
	C =.. [any,_],
	!.
ub_atom_merge(B,_,B,_) :-
	B =.. [any,_],
	!.
ub_atom_merge(C,C,C,S) :-
	C =.. [T,_],
	!,
	insert_ub_pair((T,T,T),S).
ub_atom_merge(B,C,D,More) :-
	B =.. [R,X],
	C =.. [S,X],
	get_ub_name(R,S,T,More),
	D =.. [T,X].

all_proc_names([proc(T/1,_)|Ps],S,M) :-
	memb1((T,T,T),S),
	memb1((T,T,T),M),
        all_proc_names(Ps,S,M).
all_proc_names([Qs|Ps],S,M) :-
        all_proc_names(Qs,S,M),
        all_proc_names(Ps,S,M).
all_proc_names([],_,_).

rul_defs(T,P,D1,D2) :-
	rul_def(T,P1,D1),
	rul_def(T,P2,D2),
	app(P1,P2,P).

rul_def(T/N,P,[proc(T/N,P)|_]) :-
	!.
rul_def(T,P,[proc(_,_)|Ps]) :-
	!,
        rul_def(T,P,Ps).
rul_def(_,[],[]).

rul_defs(T,P,D1,D2,D3,D4) :-
	rul_def_remove(T,P1,D1,D3),
	rul_def_remove(T,P2,D2,D4),
	app(P1,P2,P).


rul_def_remove(T/N,P,[proc(T/N,P)|Ps],Ps) :-
	!.
rul_def_remove(T,P,[proc(A,B)|Ps],[proc(A,B)|Qs]) :-
	!,
        rul_def_remove(T,P,Ps,Qs).
rul_def_remove(_,[],[],[]).


renametypes([],_,[]) :-
	!.
renametypes([(H :- B)|Ps],T,[(S :- B)|Qs]) :-
	H =.. [_,X],
	S =.. [T,X],
	renametypes(Ps,T,Qs).

get_ub_name(R,S,T,More) :-
        R @=< S,
        !,
        insert_ub_pair((R,S,T),More).
get_ub_name(R,S,T,More) :-
        insert_ub_pair((S,R,T),More).


insert_ub_pair(P,pairs(S,M)) :-
        insert_ub_pair1(P,S,M).

insert_ub_pair1((A,A,A),S,M) :-
        var(S),
        !,
        S = [(A,A,A)|_],
        memb1((A,A,A),M).
insert_ub_pair1((A,B,C),S,M) :-
        var(S),
        !,
        ub_name(A,B,C),
        S = [(A,B,C)|_],
        memb1((A,B,C),M).
insert_ub_pair1(P,[P|_],_) :-
        !.
insert_ub_pair1(P,[_|S],M) :-
        insert_ub_pair1(P,S,M).

ub_name(T1,T2,T) :-
	decompose_name(T1,D1),
        decompose_name(T2,D2),
	u_append(D1,D2,D3),
	u_normalise(D3,D4),
	compound_is_name(D4,C),
        name(T,C).

insertT(any,_,_) :-
        !.
insertT(T,S,M) :-
        var(S),
        !,
        memb1(T,M),
        S = [T|_].
insertT(T,[T|_],_) :-
        !.
insertT(T,[_|S],M) :-
        insertT(T,S,M).

/* get all clauses on which a predicate depends */

rdefs(T,Defs,P,Acc) :-
        iterate_defs([T],[T|_],Defs,[],P,Acc).

iterate_defs([],_,D,D,_,_) :-
        !.
iterate_defs(Ts,Ss,Defs,Defs2,P,Acc) :-
        each_def(Ts,Defs,Defs1,Ss,More,P,Acc),
        iterate_defs(More,Ss,Defs1,Defs2,P,Acc).

each_def([],D,D,_,_,_,_) :-
        !.
each_def([T|Ts],[proc(T/1,P)|Ps],Defs,Ss,M,D,Acc) :-
        accdef(T/1,P,D,Acc),
        body_defs(P,Ss,M),
        each_def(Ts,Ps,Defs,Ss,M,D,Acc).

body_defs([(_ :- B)|Cs],Ss,M) :-
        bodytypes(B,Ss,M),
        body_defs(Cs,Ss,M).
body_defs([],_,_).

bodytypes((B,Bs),Ss,M) :-
        !,
        functor(B,T,1),
        insertT(T,Ss,M),
        bodytypes(Bs,Ss,M).
bodytypes(true,_,_) :-
        !.
bodytypes(B,Ss,M) :-
        functor(B,T,1),
        insertT(T,Ss,M).

accdef(T/1,P,D,_) :-
	def1(T/1,P,D),
	!.
accdef(T/1,P,_,Acc) :-
	def1(T/1,P,Acc).
%	accdef2(T/1,P,Acc).

accdef2(T/1,P,[A|_]) :-
	def1(T/1,P,A),
	!.
accdef2(T/1,P,[_|As]) :-
	accdef2(T/1,P,As).

normalise_defs([],_,_,[]).
normalise_defs([proc(F,Cls)|Ps],T2,T,[proc(F,Cls1)|Ps1]) :-
        normalise(Cls,T2,T,Cls1),
        normalise_defs(Ps,T2,T,Ps1).


replaceproc(T1,[proc(T1/1,_)|Ps],Cls,[proc(T1/1,Cls)|Ps]) :-
        !.
replaceproc(T1,[P|Ps],Cls,[P|Ps1]) :-
        replaceproc(T1,Ps,Cls,Ps1).

normalise([], _,_,[]).
normalise([(H :- B)|Cls], T2,F,[(H :- B1)|T]) :-
        normalise_body(B,T2,F,B1),
        normalise(Cls,T2,F,T).

normalise_body(true,_,_,true) :-
        !.
normalise_body((B,Bs),T2,F,(B1,Bs1)) :-
        B =.. [T2,X],
        !,
        B1 =.. [F,X],
        normalise_body(Bs,T2,F,Bs1).
normalise_body((B,Bs),T2,F,(B,Bs1)) :-
        !,
        normalise_body(Bs,T2,F,Bs1).
normalise_body(B,T2,F,B1) :-
        B =.. [T2,X],
        !,
        B1 =.. [F,X].
normalise_body(B,_,_,B).

